@@ -13,10 +13,11 @@ import re
 from collections import OrderedDict
 
 import reframe.core.fields as fields
-import reframe.utility.os_ext as os_ext
+import reframe.utility.osext as osext
 import reframe.utility.typecheck as types
 from reframe.core.exceptions import (ConfigError, EnvironError,
                                      SpawnedProcessError)
+from reframe.core.logging import getlogger
 from reframe.utility import OrderedSet
 
 
@@ -27,7 +28,7 @@ class Module:
     implementation should deal only with that.
     '''
 
-    def __init__(self, name):
+    def __init__(self, name, collection=False):
         if not isinstance(name, str):
             raise TypeError('module name not a string')
 
@@ -40,6 +41,9 @@ class Module:
         except ValueError:
             self._name, self._version = name, None
 
+        # This module represents a "module collection" in TMod4
+        self._collection = collection
+
     @property
     def name(self):
         return self._name
@@ -48,6 +52,10 @@ class Module:
     def version(self):
         return self._version
 
+    @property
+    def collection(self):
+        return self._collection
+
     @property
     def fullname(self):
         if self.version is not None:
@@ -65,13 +73,16 @@ class Module:
         if not isinstance(other, type(self)):
             return NotImplemented
 
+        if self.collection != other.collection:
+            return False
+
         if not self.version or not other.version:
             return self.name == other.name
         else:
             return self.name == other.name and self.version == other.version
 
     def __repr__(self):
-        return '%s(%s)' % (type(self).__name__, self.fullname)
+        return f'{type(self).__name__}({self.fullname}, {self.collection})'
 
     def __str__(self):
         return self.fullname
@@ -85,6 +96,7 @@ class ModulesSystem:
 
     @classmethod
     def create(cls, modules_kind=None):
+        getlogger().debug(f'Initializing modules system {modules_kind!r}')
         if modules_kind is None or modules_kind == 'nomod':
             return ModulesSystem(NoModImpl())
         elif modules_kind == 'tmod31':
@@ -167,42 +179,68 @@ class ModulesSystem:
         '''
         return [str(m) for m in self._backend.loaded_modules()]
 
-    def conflicted_modules(self, name):
+    def conflicted_modules(self, name, collection=False):
         '''Return the list of the modules conflicting with module ``name``.
 
         If module ``name`` resolves to multiple real modules, then the returned
         list will be the concatenation of the conflict lists of all the real
         modules.
 
-        :rtype: List[str]
+        :arg name: The name of the module.
+        :arg collection: The module is a "module collection" (TMod4 only).
+        :returns: A list of conflicting module names.
+
+        .. versionchanged:: 3.3
+           The ``collection`` argument was added.
+
         '''
         ret = []
         for m in self.resolve_module(name):
-            ret += self._conflicted_modules(m)
+            ret += self._conflicted_modules(m, collection)
 
         return ret
 
-    def _conflicted_modules(self, name):
-        return [str(m) for m in self._backend.conflicted_modules(Module(name))]
+    def _conflicted_modules(self, name, collection=False):
+        return [
+            str(m)
+            for m in self._backend.conflicted_modules(Module(name, collection))
+        ]
 
-    def load_module(self, name, force=False):
+    def execute(self, cmd, *args):
+        '''Execute an arbitrary module command.
+
+        :arg cmd: The command to execute, e.g., ``load``, ``restore`` etc.
+        :arg args: The arguments to pass to the command.
+        :returns: The command output.
+        '''
+        return self._backend.execute(cmd, *args)
+
+    def load_module(self, name, force=False, collection=False):
         '''Load the module ``name``.
 
-        If ``force`` is set, forces the loading, unloading first any
-        conflicting modules currently loaded. If module ``name`` refers to
-        multiple real modules, all of the target modules will be loaded.
+        :arg force: If set, forces the loading, unloading first any
+            conflicting modules currently loaded. If module ``name`` refers to
+            multiple real modules, all of the target modules will be loaded.
+        :arg collection: The module is a "module collection" (TMod4 only)
+        :returns: A list of two-element tuples, where each tuple contains the
+            module that was loaded and the list of modules that had to be
+            unloaded first due to conflicts. This list will be normally of
+            size one, but it can be longer if there is mapping that maps
+            module ``name`` to multiple other modules.
+
+        .. versionchanged:: 3.3
+           - The ``collection`` argument was added.
+           - This function now returns a list of tuples.
 
-        :returns: the list of unloaded modules as strings.
-        :rtype: List[str]
         '''
         ret = []
         for m in self.resolve_module(name):
-            ret += self._load_module(m, force)
+            ret.append((m, self._load_module(m, force, collection)))
 
         return ret
 
-    def _load_module(self, name, force=False):
-        module = Module(name)
+    def _load_module(self, name, force=False, collection=False):
+        module = Module(name, collection)
         loaded_modules = self._backend.loaded_modules()
         if module in loaded_modules:
             # Do not try to load the module if it is already present
@@ -220,17 +258,23 @@ class ModulesSystem:
         self._backend.load_module(module)
         return [str(m) for m in unload_list]
 
-    def unload_module(self, name):
+    def unload_module(self, name, collection=False):
         '''Unload module ``name``.
 
-        If module ``name`` refers to multiple real modules, all the referred to
-        modules will be unloaded in reverse order.
+        :arg name: The name of the module to unload. If module ``name`` is
+            resolved to multiple real modules, all the referred to modules
+            will be unloaded in reverse order.
+        :arg collection: The module is a "module collection" (TMod4 only)
+
+        .. versionchanged:: 3.3
+           The ``collection`` argument was added.
+
         '''
         for m in reversed(self.resolve_module(name)):
-            self._unload_module(m)
+            self._unload_module(m, collection)
 
-    def _unload_module(self, name):
-        self._backend.unload_module(Module(name))
+    def _unload_module(self, name, collection=False):
+        self._backend.unload_module(Module(name, collection))
 
     def is_module_loaded(self, name):
         '''Check if module ``name`` is loaded.
@@ -309,29 +353,70 @@ class ModulesSystem:
         '''Remove ``dirs`` from the module system search path.'''
         return self._backend.searchpath_remove(*dirs)
 
-    def emit_load_commands(self, name):
-        '''Return the appropriate shell command for loading module ``name``.
+    def emit_load_commands(self, name, collection=False):
+        '''Return the appropriate shell commands for loading a module.
+
+        Module mappings are not taken into account by this function.
+
+        :arg name: The name of the module to load.
+        :arg collection: The module is a "module collection" (TMod4 only)
+        :returns: A list of shell commands.
+
+        .. versionchanged:: 3.3
+           The ``collection`` argument was added and module mappings are no
+           more taken into account by this function.
 
-        :rtype: List[str]
         '''
-        return [self._backend.emit_load_instr(Module(name))
-                for name in self.resolve_module(name)]
 
-    def emit_unload_commands(self, name):
-        '''Return the appropriate shell command for unloading module
-        ``name``.
+        # We don't consider module mappings here, because we cannot treat
+        # correctly possible conflicts
+        return [self._backend.emit_load_instr(Module(name, collection))]
+
+    def emit_unload_commands(self, name, collection=False):
+        '''Return the appropriate shell commands for unloading a module.
+
+        Module mappings are not taken into account by this function.
+
+        :arg name: The name of the module to unload.
+        :arg collection: The module is a "module collection" (TMod4 only)
+        :returns: A list of shell commands.
+
+        .. versionchanged:: 3.3
+           The ``collection`` argument was added and module mappings are no
+           more taken into account by this function.
 
-        :rtype: List[str]
         '''
-        return [self._backend.emit_unload_instr(Module(name))
-                for name in reversed(self.resolve_module(name))]
+
+        # See comment in emit_load_commands()
+        return [self._backend.emit_unload_instr(Module(name, collection))]
 
     def __str__(self):
         return str(self._backend)
 
 
 class ModulesSystemImpl(abc.ABC):
-    '''Abstract base class for module systems.'''
+    '''Abstract base class for module systems.
+
+    :meta private:
+    '''
+
+    def execute(self, cmd, *args):
+        '''Execute an arbitrary module command using the modules backend.
+
+        :arg cmd: The command to execute, e.g., ``load``, ``restore`` etc.
+        :arg args: The arguments to pass to the command.
+        :returns: The command output.
+        '''
+        try:
+            exec_output = self._execute(cmd, *args)
+        except SpawnedProcessError as e:
+            raise EnvironError('could not execute module operation') from e
+
+        return exec_output
+
+    @abc.abstractmethod
+    def _execute(self, cmd, *args):
+        '''Execute an arbitrary command of the module system.'''
 
     @abc.abstractmethod
     def available_modules(self, substr):
@@ -379,6 +464,10 @@ class ModulesSystemImpl(abc.ABC):
     def version(self):
         '''Return the version of this module system.'''
 
+    @abc.abstractmethod
+    def modulecmd(self, *args):
+        '''The low level command to use for issuing module commads'''
+
     @abc.abstractmethod
     def unload_all(self):
         '''Unload all loaded modules.'''
@@ -418,7 +507,7 @@ class TModImpl(ModulesSystemImpl):
     def __init__(self):
         # Try to figure out if we are indeed using the TCL version
         try:
-            completed = os_ext.run_command('modulecmd -V')
+            completed = osext.run_command('modulecmd -V')
         except OSError as e:
             raise ConfigError(
                 'could not find a sane TMod installation') from e
@@ -433,7 +522,7 @@ class TModImpl(ModulesSystemImpl):
 
         version = version_match.group(1)
         try:
-            ver_major, ver_minor, *_ = [int(v) for v in version.split('.')]
+            ver_major, ver_minor = [int(v) for v in version.split('.')[:2]]
         except ValueError:
             raise ConfigError(
                 'could not parse TMod version string: ' + version) from None
@@ -444,10 +533,9 @@ class TModImpl(ModulesSystemImpl):
                 (version, self.MIN_VERSION))
 
         self._version = version
-        self._command = 'modulecmd python'
         try:
             # Try the Python bindings now
-            completed = os_ext.run_command(self._command)
+            completed = osext.run_command(self.modulecmd())
         except OSError as e:
             raise ConfigError(
                 'could not get the Python bindings for TMod: ' % e) from e
@@ -462,35 +550,25 @@ class TModImpl(ModulesSystemImpl):
     def version(self):
         return self._version
 
-    def _run_module_command(self, *args, msg=None):
-        command = ' '.join([self._command, *args])
-        try:
-            completed = os_ext.run_command(command, check=True)
-        except SpawnedProcessError as e:
-            raise EnvironError(msg) from e
+    def modulecmd(self, *args):
+        return ' '.join(['modulecmd', 'python', *args])
 
-        if self._module_command_failed(completed):
-            err = SpawnedProcessError(command,
+    def _execute(self, cmd, *args):
+        modulecmd = self.modulecmd(cmd, *args)
+        completed = osext.run_command(modulecmd)
+        if re.search(r'ERROR', completed.stderr) is not None:
+            raise SpawnedProcessError(modulecmd,
                                       completed.stdout,
                                       completed.stderr,
                                       completed.returncode)
-            raise EnvironError(msg) from err
 
-        return completed
-
-    def _module_command_failed(self, completed):
-        return re.search(r'ERROR', completed.stderr) is not None
-
-    def _exec_module_command(self, *args, msg=None):
-        completed = self._run_module_command(*args, msg=msg)
         exec(completed.stdout)
+        return completed.stderr
 
     def available_modules(self, substr):
-        completed = self._run_module_command(
-            'avail', '-t', substr, msg='could not retrieve available modules'
-        )
+        output = self.execute('avail', '-t', substr)
         ret = []
-        for line in completed.stderr.split('\n'):
+        for line in output.split('\n'):
             if not line or line[-1] == ':':
                 # Ignore empty lines and path entries
                 continue
@@ -509,43 +587,38 @@ class TModImpl(ModulesSystemImpl):
             return []
 
     def conflicted_modules(self, module):
-        completed = self._run_module_command(
-            'show', str(module), msg="could not show module '%s'" % module)
+        output = self.execute('show', str(module))
         return [Module(m.group(1))
                 for m in re.finditer(r'^conflict\s+(\S+)',
-                                     completed.stderr, re.MULTILINE)]
+                                     output, re.MULTILINE)]
 
     def is_module_loaded(self, module):
         return module in self.loaded_modules()
 
     def load_module(self, module):
-        self._exec_module_command(
-            'load', str(module),
-            msg="could not load module '%s' correctly" % module)
+        self.execute('load', str(module))
 
     def unload_module(self, module):
-        self._exec_module_command(
-            'unload', str(module),
-            msg="could not unload module '%s' correctly" % module)
+        self.execute('unload', str(module))
 
     def unload_all(self):
-        self._exec_module_command('purge')
+        self.execute('purge')
 
     def searchpath(self):
         path = os.getenv('MODULEPATH', '')
         return path.split(':')
 
     def searchpath_add(self, *dirs):
-        self._exec_module_command('use', *dirs)
+        self.execute('use', *dirs)
 
     def searchpath_remove(self, *dirs):
-        self._exec_module_command('unuse', *dirs)
+        self.execute('unuse', *dirs)
 
     def emit_load_instr(self, module):
-        return 'module load %s' % module
+        return f'module load {module}'
 
     def emit_unload_instr(self, module):
-        return 'module unload %s' % module
+        return f'module unload {module}'
 
 
 class TMod31Impl(TModImpl):
@@ -558,7 +631,7 @@ class TMod31Impl(TModImpl):
         try:
             modulecmd = os.getenv('MODULESHOME')
             modulecmd = os.path.join(modulecmd, 'modulecmd.tcl')
-            completed = os_ext.run_command(modulecmd)
+            completed = osext.run_command(modulecmd)
         except OSError as e:
             raise ConfigError(
                 'could not find a sane TMod31 installation: %s' % e) from e
@@ -572,7 +645,7 @@ class TMod31Impl(TModImpl):
 
         version = version_match.group(1)
         try:
-            ver_major, ver_minor, *_ = [int(v) for v in version.split('.')]
+            ver_major, ver_minor = [int(v) for v in version.split('.')[:2]]
         except ValueError:
             raise ConfigError(
                 'could not parse TMod31 version string: ' + version) from None
@@ -587,7 +660,7 @@ class TMod31Impl(TModImpl):
 
         try:
             # Try the Python bindings now
-            completed = os_ext.run_command(self._command)
+            completed = osext.run_command(self._command)
         except OSError as e:
             raise ConfigError(
                 'could not get the Python bindings for TMod31: ' % e) from e
@@ -599,21 +672,28 @@ class TMod31Impl(TModImpl):
     def name(self):
         return 'tmod31'
 
-    def _exec_module_command(self, *args, msg=None):
-        completed = self._run_module_command(*args, msg=msg)
-        exec_match = re.search(r'^exec\s\'', completed.stdout)
+    def modulecmd(self, *args):
+        return ' '.join([self._command, *args])
+
+    def _execute(self, cmd, *args):
+        modulecmd = self.modulecmd(cmd, *args)
+        completed = osext.run_command(modulecmd)
+        if re.search(r'ERROR', completed.stderr) is not None:
+            raise SpawnedProcessError(modulecmd,
+                                      completed.stdout,
+                                      completed.stderr,
+                                      completed.returncode)
+
+        exec_match = re.search(r"^exec\s'(\S+)'", completed.stdout,
+                               re.MULTILINE)
         if exec_match is None:
             raise ConfigError('could not use the python bindings')
-        else:
-            cmd = completed.stdout
-            exec_match = re.search(r'^exec\s\'(\S+)\'', cmd,
-                                   re.MULTILINE)
-            if exec_match is None:
-                raise ConfigError('could not use the python bindings')
-            with open(exec_match.group(1), 'r') as content_file:
-                cmd = content_file.read()
+
+        with open(exec_match.group(1), 'r') as content_file:
+            cmd = content_file.read()
 
         exec(cmd)
+        return completed.stderr
 
 
 class TMod4Impl(TModImpl):
@@ -622,9 +702,8 @@ class TMod4Impl(TModImpl):
     MIN_VERSION = (4, 1)
 
     def __init__(self):
-        self._command = 'modulecmd python'
         try:
-            completed = os_ext.run_command(self._command + ' -V', check=True)
+            completed = osext.run_command(self.modulecmd('-V'), check=True)
         except OSError as e:
             raise ConfigError(
                 'could not find a sane TMod4 installation') from e
@@ -639,7 +718,7 @@ class TMod4Impl(TModImpl):
 
         version = version_match.group(1)
         try:
-            ver_major, ver_minor, *_ = [int(v) for v in version.split('.')]
+            ver_major, ver_minor = [int(v) for v in version.split('.')[:2]]
         except ValueError:
             raise ConfigError(
                 'could not parse TMod4 version string: ' + version) from None
@@ -654,35 +733,73 @@ class TMod4Impl(TModImpl):
     def name(self):
         return 'tmod4'
 
-    def _exec_module_command(self, *args, msg=None):
-        command = ' '.join([self._command, *args])
-        completed = os_ext.run_command(command, check=True)
+    def modulecmd(self, *args):
+        return ' '.join(['modulecmd', 'python', *args])
+
+    def _execute(self, cmd, *args):
+        modulecmd = self.modulecmd(cmd, *args)
+        completed = osext.run_command(modulecmd, check=False)
         namespace = {}
         exec(completed.stdout, {}, namespace)
-        if not namespace['_mlstatus']:
-            # _mlstatus is set by the TMod4 Python bindings
-            if msg is None:
-                msg = 'modules system command failed: '
-                if isinstance(completed.args, str):
-                    msg += completed.args
-                else:
-                    msg += ' '.join(completed.args)
 
-            raise EnvironError(msg)
+        # _mlstatus is set by the TMod4 only if the command was unsuccessful,
+        # but Lmod sets it always
+        if not namespace.get('_mlstatus', True):
+            raise SpawnedProcessError(modulecmd,
+                                      completed.stdout,
+                                      completed.stderr,
+                                      completed.returncode)
+
+        return completed.stderr
+
+    def load_module(self, module):
+        if module.collection:
+            self.execute('restore', str(module))
+            return []
+        else:
+            return super().load_module(module)
+
+    def unload_module(self, module):
+        if module.collection:
+            # Module collection are not unloaded
+            return
+
+        super().unload_module(module)
+
+    def conflicted_modules(self, module):
+        if module.collection:
+            # Conflicts have no meaning in module collection. The modules
+            # system will take care of these when restoring a module
+            # collection
+            return []
+
+        return super().conflicted_modules(module)
+
+    def emit_load_instr(self, module):
+        if module.collection:
+            return f'module restore {module}'
+
+        return super().emit_load_instr(module)
+
+    def emit_unload_instr(self, module):
+        if module.collection:
+            return ''
+
+        return super().emit_unload_instr(module)
 
 
-class LModImpl(TModImpl):
+class LModImpl(TMod4Impl):
     '''Module system for Lmod (Tcl/Lua).'''
 
     def __init__(self):
         # Try to figure out if we are indeed using LMOD
-        lmod_cmd = os.getenv('LMOD_CMD')
-        if lmod_cmd is None:
+        self._lmod_cmd = os.getenv('LMOD_CMD')
+        if self._lmod_cmd is None:
             raise ConfigError('could not find a sane Lmod installation: '
                               'environment variable LMOD_CMD is not defined')
 
         try:
-            completed = os_ext.run_command('%s --version' % lmod_cmd)
+            completed = osext.run_command(f'{self._lmod_cmd} --version')
         except OSError as e:
             raise ConfigError(
                 'could not find a sane Lmod installation: %s' % e)
@@ -693,10 +810,9 @@ class LModImpl(TModImpl):
             raise ConfigError('could not retrieve Lmod version')
 
         self._version = version_match.group(1)
-        self._command = '%s python ' % lmod_cmd
         try:
             # Try the Python bindings now
-            completed = os_ext.run_command(self._command)
+            completed = osext.run_command(self.modulecmd())
         except OSError as e:
             raise ConfigError(
                 'could not get the Python bindings for Lmod: ' % e)
@@ -708,15 +824,13 @@ class LModImpl(TModImpl):
     def name(self):
         return 'lmod'
 
-    def _module_command_failed(self, completed):
-        return completed.stdout.strip() == 'false'
+    def modulecmd(self, *args):
+        return ' '.join([self._lmod_cmd, 'python', *args])
 
     def available_modules(self, substr):
-        completed = self._run_module_command(
-            '-t', 'avail', substr, msg='could not retrieve available modules'
-        )
+        output = self.execute('-t', 'avail', substr)
         ret = []
-        for line in completed.stderr.split('\n'):
+        for line in output.split('\n'):
             if not line or line[-1] == ':':
                 # Ignore empty lines and path entries
                 continue
@@ -727,15 +841,20 @@ class LModImpl(TModImpl):
         return ret
 
     def conflicted_modules(self, module):
-        completed = self._run_module_command(
-            'show', str(module), msg="could not show module '%s'" % module)
+        if module.collection:
+            # Conflicts have no meaning in module collection. The modules
+            # system will take care of these when restoring a module
+            # collection
+            return []
+
+        output = self.execute('show', str(module))
 
         # Lmod accepts both Lua and and Tcl syntax
         # The following test allows incorrect syntax, e.g., `conflict
         # ('package"(`, but we expect this to be caught by the Lmod framework
         # in earlier stages.
         ret = []
-        for m in re.finditer(r'conflict\s*(\S+)', completed.stderr):
+        for m in re.finditer(r'conflict\s*(\S+)', output):
             conflict_arg = m.group(1)
             if conflict_arg.startswith('('):
                 # Lua syntax
@@ -746,10 +865,24 @@ class LModImpl(TModImpl):
 
         return ret
 
+    def load_module(self, module):
+        if module.collection:
+            self.execute('restore', str(module))
+            return []
+        else:
+            return super().load_module(module)
+
+    def unload_module(self, module):
+        if module.collection:
+            # Module collection are not unloaded
+            return
+
+        super().unload_module(module)
+
     def unload_all(self):
         # Currently, we don't take any provision for sticky modules in Lmod, so
         # we forcefully unload everything.
-        self._exec_module_command('--force', 'purge')
+        self.execute('--force', 'purge')
 
 
 class NoModImpl(ModulesSystemImpl):
@@ -764,6 +897,9 @@ class NoModImpl(ModulesSystemImpl):
     def conflicted_modules(self, module):
         return []
 
+    def _execute(self, cmd, *args):
+        return ''
+
     def load_module(self, module):
         pass
 
@@ -783,6 +919,9 @@ class NoModImpl(ModulesSystemImpl):
     def version(self):
         return '1.0'
 
+    def modulecmd(self, *args):
+        return ''
+
     def unload_all(self):
         pass
 
